#! /usr/bin/env python
"""
Copyright (c) 2019 - 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import os
import cv2
import ast
import sys
import time
import glob
import argparse
import subprocess

from threading import Thread
from multiprocess import Pool
from pythutils.sysutils import lineprint
from pythutils.drawutils import draw_text
from pythutils.fileutils import listfiles, get_ext, commonpref, move
from pythutils.mediautils import get_vid_params, videowriter, imgresize

class KeyboardInterruptError(Exception): pass

class Convert:

    """
    Module to convert a directory of media files with potential to resize the
    media, write the unique frame number on each frame, and continuously monitor
    a folder for updated files. Multiple files can be converted simultaneously
    with the pools parameter to optimally use the computer's processing cores.

    Parameters
    -----------
    indir : str, default = ""
        Directory containing the videos.
    outdir : str, default = ""
        Directory where the converted videos should be stored. If the directory
        does not exist yet it will be newly created.
    type : str, default = ".h264"
        The filetype of the media to convert.
    withframe : bool, default = False
        Type of conversion, either very fast conversion using FFmpeg or using
        OpenCV to draw the frame number on each video frame.
    delete : bool, default = False
        If the original videos should be deleted or not.
    pools : int, default = 4
        Number of simultaneous converting processing that should be allowed.
        Works optimally when equal to the number of computer processing cores.
    resizeval : float, default = 1
        Float value to which the video should be resized.
    imgfps : int, default = 25
        Framerate for conversion of images to video.
    sleeptime : int, default = None
        Time in seconds between subsequent checks for files within a folder. The
        default value (None) only converts the current files.
    """

    def __init__(self, indir = "", outdir = "", type = ".h264",
                 withframe = False, overwrite = False, delete = False,
                 pools = 4, resizeval = 1, fps = None, imgfps = 25,
                 internal = False, sleeptime = None):

        if internal:
            lineprint("Running convert function..", label="pirecorder")

        self.indir = os.getcwd() if indir == "" else indir
        assert os.path.exists(self.indir), "in-directory does not exist.."
        self.outdir = self.indir if outdir == "" else outdir
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)

        self.rootdir = os.getcwd()
        os.chdir(self.indir)

        self.type = type
        self.withframe = withframe
        self.delete = delete
        self.pools = int(pools)
        self.resizeval = float(resizeval)
        self.fps = int(fps) if fps is not None else None
        self.imgfps = int(imgfps)
        self.terminated = False

        while True:
            files = listfiles(self.indir, self.type, keepdir = False)
            old = listfiles(self.indir, self.type, keepext = False)
            new = listfiles(self.outdir, ".mp4", keepext = False)
            self.todo = files
            if not overwrite:
                self.todo = [files[i] for i,file in enumerate(old) if file not in new]
            if self.type in [".jpg",".jpeg",".png"] and len(self.todo)>0:
                 if len([f for f in new if commonpref(self.todo) in f])>0 and not overwrite:
                     self.todo = []
            self.convertpool()
            msg = "No files to convert.."
            if sleeptime == None:
                if not self.terminated:
                    lineprint(msg, label="pirecorder")
                return
            else:
                try:
                    lineprint(msg+" rechecking in "+str(sleeptime)+"s..", label="pirecorder")
                    time.sleep(sleeptime)
                except KeyboardInterrupt:
                    lineprint("Terminating checking for files..", label="pirecorder")
                    return

    def conv_single(self, filein):

        try:
            filebase = os.path.basename(filein)
            fileout = filein if self.outdir == "" else self.outdir+"/"+filebase
            lineprint("Start converting "+filebase, label="pirecorder")

            if self.withframe:
                vid = cv2.VideoCapture(filein)
                fps, width, height, _ = get_vid_params(vid)
                if self.fps is None:
                    self.fps = fps
                vidout = videowriter(fileout, width, height, self.fps, self.resizeval)

                while True:
                    flag, frame = vid.read()
                    if flag:
                        if self.resizeval != 1:
                            frame = imgresize(frame, self.resizeval)
                        frame_nr = int(vid.get(cv2.CAP_PROP_POS_FRAMES))
                        draw_text(frame, str(frame_nr), (10,10), 0.9, col="white",
                                  shadow=True)
                        vidout.write(frame)
                    if not flag:
                        break

            else:
                if self.resizeval != 1:
                    comm = "' -vf 'scale=iw*" + str(self.resizeval) + ":-2' '"
                else:
                    comm = "' -vcodec copy '"
                bashcomm = "ffmpeg"
                if self.fps is not None:
                    bashcomm = bashcomm+" -r "+ str(self.fps)
                bashcomm = bashcomm+" -i '"+filein+comm+fileout[:-len(self.type)]+".mp4'"
                bashcomm = bashcomm + " -y -nostats -loglevel 0"
                output = subprocess.check_output(['bash','-c', bashcomm])

            lineprint("Finished converting "+filebase, label="pirecorder")

        except KeyboardInterrupt:
            raise KeyboardInterruptError()


    def convertpool(self):

        if len(self.todo) > 0:

            if self.type in [".h264",".mp4",".avi"]:

                pool = Pool(min(self.pools, len(self.todo)))
                try:
                    pool.map(self.conv_single, self.todo)
                    pool.close()
                    lineprint("Done converting all videofiles!", label="pirecorder")
                except KeyboardInterrupt:
                    lineprint("User terminated converting pool..", label="pirecorder")
                    self.terminated = True
                    pool.terminate()
                    return
                except Exception as e:
                    excep = "Got exception: %r, terminating pool" % (e,)
                    lineprint(excep, label="pirecorder")
                    pool.terminate()
                finally:
                    pool.join()

                if self.delete:
                    for filein in self.todo:
                        os.remove(filein)
                    lineprint("Deleted all original videofiles..", label="pirecorder")

            elif self.type in [".jpg",".jpeg",".png"]:

                vidname = commonpref(self.todo)
                lineprint("Start converting "+str(len(self.todo))+" images", label="pirecorder")

                frame_array = []
                for filename in self.todo:
                    frame = cv2.imread(filename)
                    frame_array.append(frame)
                    #os.rename(filename, self.outdir+"/"+filename)
                h, w, _ = frame_array[0].shape
                if self.outdir != "":
                    vidname = self.outdir+"/"+os.path.basename(vidname)
                vidout = videowriter(vidname, w, h, self.imgfps, self.resizeval)
                for i in range(len(frame_array)):
                    vidout.write(frame_array[i])
                vidout.release()
                lineprint("Finished converting "+os.path.basename(vidname), label="pirecorder")

            else:
                lineprint("No video or image files found..", label="pirecorder")


def conv():

    """To run the convert function from the command line"""

    parser = argparse.ArgumentParser(prog="Convert", description=Convert.__doc__,
             formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("-i", "--indir" ,default="", metavar="")
    parser.add_argument("-o", "--outdir" ,default="", metavar="")
    parser.add_argument("-t", "--type", default=".h264", metavar="")
    parser.add_argument("-w", "--withframe", default="False", metavar="")
    parser.add_argument("-d", "--delete", default="False", metavar="")
    parser.add_argument("-p", "--pools", default=4, type=int, metavar="")
    parser.add_argument("-r", "--resizeval", default=1, type=float, metavar="")
    parser.add_argument("-f", "--imgfps", default=25, type=int, metavar="")
    parser.add_argument("-s", "--sleeptime", default=None, type=int, metavar="")

    args = parser.parse_args()
    args.withframe = ast.literal_eval(args.withframe)
    args.delete = ast.literal_eval(args.delete)
    Convert(indir = args.indir, outdir = args.outdir, type = args.type,
            withframe = args.withframe, delete = args.delete, pools = args.pools,
            resizeval = args.resizeval, imgfps = args.imgfps,
            sleeptime = args.sleeptime)



#! /usr/bin/env python
"""
Copyright (c) 2019 - 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import cv2
import time
import argparse
import numpy as np

from pythutils.sysutils import lineprint
from pythutils.mediautils import checkroi, roi_to_zoom, imgresize
import pythutils.drawutils as draw

from .videoin import VideoIn
from .__version__ import __version__

class Stream:

    def __init__(self, system = "auto", framerate = 8, vidsize = 0.2,
                 rotation = 0, internal = False, maxres = None,
                 imgoverlay = None):

        """
        Opens a video stream with user interface to help position and
        adjust the camera

        parameters
        -----------
        system : str, default = "auto"
            If the system should be automatically determined. Should detect if
            the computer is a raspberry pi or not. If this somehow fails, set
            to "rpi" manually.
        framerate : int, default = 8
            The framerate of the displayed video stream. Lower framerates take
            longer to start up. When using an image overlay, maximum possible
            framerate is 5, to avoid getting shutter effects
        vidsize : float, default = 0.2
            The relative size of the video window to the maximum resolution of
            the raspberry pi camera type.
        rotation : int, default = 180
            If the camera should be rotated or not. 0 and 180 are valid values.
        maxres : str or tuple, default = "v2"
            The maximum potential resolution of the camera used. Either provide
            a tuple of the max resolution, or use "v1.5", "v2" (default), or
            "hq" to get the maximum resolution associated with the official cameras
            directly.
        imgoverlay : str, default = None
            The path to an image that will be overlaid on the video stream. This
            can be helpful for accurate positioning of the camera in line with
            previous recordings or setups.

        interface
        -----------
        This interactive module stores mouse position and clicks and responds to
        the following keypresses:
        f-key : display the stream fullscreen/display the stream in a window
        c-key : display/hide a diagonal cross across the screen
        s-key : save the coordinates of the rectangular area when drawn
        e-key : erase the rectangular area when drawn
        z-key : show a zoomed-in section of the video inside the rectangular
            area in maximum resolution
        n-key : refresh the zoom-in image
        o-key : if the potential overlay image should be shown or not
        [- and ]-keys : decrease or increase the relative opacity of the
            potential overlay image with 5%
        esc-key : exit the the zoom window; exit the calibrate function
        """

        self.internal = internal
        if self.internal:
            lineprint("Running stream function.. ")

        self.system = system
        self.framerate = framerate
        self.vidsize = vidsize
        self.rotation = rotation
        self.maxres = maxres

        if imgoverlay == None:
            self.overlay = False
            self.waitms = 1
        else:
            if os.path.isfile(imgoverlay):
                self.overlay = True
                self.overlayimg = cv2.imread(imgoverlay)
                self.alpha = 0.5
                self.waitms = 200
                self.framerate = min(self.framerate, 5)
            else:
                print("Image file could not be loaded..")

        self.cross = False
        self.stream = True
        self.exit = False
        self.roi = False
        self.fullscreen = False
        self.tempcol = draw.namedcols("orange")
        self.col = draw.namedcols("red")

        cv2.namedWindow("Image", cv2.WND_PROP_FULLSCREEN)
        self.m = draw.mouse_events()
        cv2.setMouseCallback('Image', self.m.draw)
        time.sleep(1)

        self.drawer()


    def draw_stream(self):

        lineprint("Streaming video..")

        self.vid = VideoIn(system=self.system, framerate=self.framerate,
                           vidsize=self.vidsize, rotation=self.rotation,
                           maxres=self.maxres)
        self.vid.start()

        if self.overlay:
            h, w, _ = self.vid.read().shape
            self.overlayimg = imgresize(self.overlayimg, resize=1, dims=(w,h))

        while True:
            self.img = self.vid.read()

            if self.overlay:
                overlay = self.img.copy()
                overlay[0:h,0:w] = self.overlayimg
                cv2.addWeighted(overlay, self.alpha, self.img, 1-self.alpha, 0, self.img)
            if self.cross:
                draw.draw_cross(self.img, pt2 = self.vid.res)
            if self.m.twoPoint is not None:
                draw.draw_crosshair(self.img, self.m.pos)
            if self.m.posDown is not None:
                cv2.rectangle(self.img, self.m.posDown, self.m.pos,
                              self.tempcol, 2)
            if self.m.posUp is not None:
                cv2.rectangle(self.img, self.m.pts[-1], self.m.posUp,
                              self.col, 2)
            cv2.imshow("Image", self.img)

            k = cv2.waitKey(self.waitms) & 0xFF
            if k == ord("o"):
                self.overlay = not self.overlay
            if k == ord("c"):
                self.cross = not self.cross
            if k == ord("f"):
                self.fullscreen = not self.fullscreen
                if self.fullscreen:
                    cv2.setWindowProperty("Image", cv2.WND_PROP_FULLSCREEN,
                                          cv2.WINDOW_FULLSCREEN)
                else:
                    cv2.setWindowProperty("Image", cv2.WND_PROP_AUTOSIZE,
                                          cv2.WINDOW_NORMAL)
                    cv2.resizeWindow("Image", self.vid.res[0], self.vid.res[1])
            if k == ord("s"):
                if self.m.twoPoint is not None:
                    self.m.twoPoint = checkroi(self.m.twoPoint, self.vid.res)
                    self.roi = roi_to_zoom(self.m.twoPoint, self.vid.res)
                    if self.internal is True:
                        lineprint("roi "+str(self.roi)+" stored..")
                    else:
                        lineprint("roi: "+str(self.roi))
                else:
                    lineprint("Nothing selected..")
            if k == ord("e"):
                self.m.posUp = None
                self.roi = False
                lineprint("roi data erased..")
            if k == ord("z"):
                if self.m.twoPoint is not None:
                    lineprint("Creating zoomed image..")
                    self.stream = False
                    break
            if self.overlay and k == ord("["):
                self.alpha = max(self.alpha-0.05, 0)
            if self.overlay and k == ord("]"):
                self.alpha = min(self.alpha+0.05, 1)
            if k == 27:
                lineprint("User exited..")
                self.exit = True
                break

        self.vid.stop()
        time.sleep(1)


    def drawer(self):
        while True:
            if self.stream:
                self.draw_stream()

            if not self.stream:
                cv2.namedWindow("Zoomed", cv2.WINDOW_NORMAL)
                cv2.setWindowProperty("Zoomed", cv2.WND_PROP_AUTOSIZE,
                                      cv2.WINDOW_NORMAL)
                while True:
                    self.vid2 = VideoIn(system=self.system, vidsize=self.vidsize,
                                        crop=self.m.twoPoint, rotation=self.rotation,
                                        maxres=self.maxres)
                    zimg = self.vid2.img()
                    cv2.imshow("Zoomed", zimg)
                    cv2.resizeWindow("Zoomed", self.vid2.roiw, self.vid2.roih)
                    k = cv2.waitKey(0) & 0xFF
                    if k == 27:
                        break
                k = 255
                cv2.destroyWindow("Zoomed")
                self.stream = True

            if self.exit:
                cv2.waitKey(1)
                cv2.destroyAllWindows()
                for i in range(5):
                    cv2.waitKey(1)
                break

def strm():

    """To run the stream function from the command line"""

    parser = argparse.ArgumentParser(prog="Stream",
             description=Stream.__doc__,
             formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("-f", "--framerate", default=8, type=int, metavar="")
    parser.add_argument("-v", "--vidsize", default=0.2, type=float, metavar="")
    parser.add_argument("-r", "--rotation", default=0, type=int, metavar="")
    parser.add_argument("-o", "--imgoverlay", default=None, metavar="")
    parser.add_argument("-m", "--maxres", default="v2", metavar="")
    parser.add_argument("-c", "--configfile", default=None,
                        action="store", help="pirecorder configuration file")

    args = parser.parse_args()
    if args.configfile is None:
        Stream(framerate = args.framerate, vidsize = args.vidsize,
               rotation = args.rotation, maxres = args.maxres,
               imgoverlay = args.imgoverlay)
    else:
        from pirecorder import PiRecorder
        rec = PiRecorder(args.configfile)
        rec.settings(internal = True)
        rec.stream()



#! /usr/bin/env python
"""
Copyright (c) 2020 Jolle Jolles <j.w.jolles@gmail.com>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at:

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import cv2
import time
import argparse

from pythutils.mediautils import picamconv
from pythutils.mathutils import maxrect
from pythutils.sysutils import checkfrac, isrpi, lineprint

def Camconfig(cam = None, auto = None, iso = 200, framerate = 20,
              res = (1640, 1232), vidsize = 0.4):

    """
    Opens a video stream to configure a wide array of camera parameters
    Note: A screen resolution of at least 800x600 is strongly recommended
    """

    if not isrpi():
        lineprint("PiRecorder only works on a raspberry pi. Exiting..")
        return

    import picamera
    import picamera.array

    def nothing(x):
        pass

    if cam == None:
        cam = picamera.PiCamera()
        cam.resolution = res
        cam.iso = iso
        cam.framerate = framerate
        time.sleep(2)
    if cam == None or (auto == None or auto == True):
        cam.exposure_mode = "auto"
        set_auto = 1
    else:
        set_auto = 0

    res = (cam.resolution[0]*vidsize, cam.resolution[1]*vidsize)
    res = (int(res[0]*cam.zoom[2]),int(res[1]*cam.zoom[3]))
    if (res[0]*res[1])>=(1640*1232):
        res = maxrect(dims, maxdims = (1640,1232), decimals = 0)
    cam.resolution = picamconv(res)

    cv2.namedWindow("Stream", cv2.WND_PROP_FULLSCREEN)
    cv2.namedWindow("Config", cv2.WINDOW_NORMAL)
    cv2.setWindowProperty("Stream", cv2.WND_PROP_AUTOSIZE, cv2.WINDOW_NORMAL)
    cv2.resizeWindow("Stream", cam.resolution[0], cam.resolution[1])

    config = []
    isos = [100,200,320,400,500,640,800]

    set_rot  = cam.rotation*180
    set_fps  = int(cam.framerate)
    set_iso  = [i for i,j in enumerate(isos) if j==cam.iso][0]
    set_comp = cam.exposure_compensation+25
    set_shut = cam.shutter_speed
    set_red  = int(float(cam.awb_gains[0])*10)
    set_blue = int(float(cam.awb_gains[1])*10)
    set_bri  = cam.brightness
    set_con  = cam.contrast+100
    set_sat  = cam.saturation+100
    set_shar = cam.sharpness+100

    cv2.createTrackbar("rotation (0deg/180deg)", "Config", set_rot, 1, nothing)
    cv2.createTrackbar("framerate", "Config", set_fps, 40, nothing)
    cv2.createTrackbar("automatic (off/on)", "Config", set_auto, 1, nothing)
    cv2.createTrackbar("iso", "Config", set_iso, 6, nothing)
    cv2.createTrackbar("compensation", "Config", set_comp, 50, nothing)
    cv2.createTrackbar("shutterspeed (ms)", "Config", set_shut, 99999, nothing)
    cv2.createTrackbar("red gain", "Config", set_red, 80, nothing)
    cv2.createTrackbar("blue gain", "Config", set_blue, 80, nothing)
    cv2.createTrackbar("brightness", "Config", set_bri, 100, nothing)
    cv2.createTrackbar("contrast", "Config", set_con, 200, nothing)
    cv2.createTrackbar("saturation", "Config", set_sat, 200, nothing)
    cv2.createTrackbar("sharpness", "Config", set_shar, 200, nothing)

    rawCapture = picamera.array.PiRGBArray(cam, size=cam.resolution)
    lineprint("Streaming interactive video..")
    with rawCapture as stream:
        while True:
            cam.capture(stream, format="bgr", use_video_port=True)
            image = stream.array

            rot  = cv2.getTrackbarPos("rotation (0deg/180deg)", "Config")
            fps  = cv2.getTrackbarPos("framerate", "Config")
            auto = cv2.getTrackbarPos("automatic (off/on)", "Config")
            iso  = cv2.getTrackbarPos("iso", "Config")
            comp = cv2.getTrackbarPos("compensation", "Config")
            shut = cv2.getTrackbarPos("shutterspeed (ms)", "Config")
            red  = cv2.getTrackbarPos("red gain", "Config")
            blue = cv2.getTrackbarPos("blue gain", "Config")
            bri  = cv2.getTrackbarPos("brightness", "Config")
            con  = cv2.getTrackbarPos("contrast", "Config")
            sat  = cv2.getTrackbarPos("saturation", "Config")
            shar = cv2.getTrackbarPos("sharpness", "Config")

            cam.rotation = [0,180][rot]
            cam.framerate = max(fps,1)
            cam.exposure_mode = ["off","auto"][auto]
            cam.iso = isos[iso]
            cam.exposure_compensation = comp-25
            cam.awb_mode = cam.exposure_mode

            if cam.exposure_mode == "off":
                cam.shutter_speed = shut
                maxshut = int(float(1/cam.framerate)*1000000)
                if shut > maxshut:
                    cv2.setTrackbarPos("shutterspeed (ms)","Config", maxshut)
                cam.awb_gains = (red/10, blue/10)

            if cam.exposure_mode == "auto":
                cam.shutter_speed = 0
                shut = cam.exposure_speed
                red, blue = [int(float(i)*10) for i in cam.awb_gains]
                cv2.setTrackbarPos("shutterspeed (ms)", "Config", shut)
                cv2.setTrackbarPos("red gain", "Config", red)
                cv2.setTrackbarPos("blue gain", "Config", blue)

            cam.brightness = bri
            cam.contrast = con-100
            cam.saturation = sat-100
            cam.sharpness = shar-100

            cv2.imshow("Stream", image)
            stream.truncate(0)

            k = cv2.waitKey(10) & 0xFF
            if k == ord("s"):
                if cam.shutter_speed == 0:
                    shutterspeed = cam.exposure_speed
                else:
                    shutterspeed = cam.shutter_speed
                gains = tuple([round(float(i),2) for i in cam.awb_gains])
                config = {"automode": [False,True][auto],
                          "vidfps": cam.framerate,
                          "rotation": cam.rotation,
                          "gains": gains,
                          "brightness": cam.brightness,
                          "contrast": cam.contrast,
                          "saturation": cam.saturation,
                          "iso":cam.iso,
                          "sharpness": cam.sharpness,
                          "compensation": cam.exposure_compensation,
                          "shutterspeed": shutterspeed}
                break
            if k == 27:
                lineprint("User exited..")
                break

        rawCapture.close()
        cam.close()
        cv2.destroyAllWindows()
        cv2.waitKey(1)

        return config


def config():

    """To run the camconfig function from the command line"""

    parser = argparse.ArgumentParser(prog="Camconfig",
             description=Camconfig.__doc__,
             formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("-a","--auto", default=None, type=str, metavar="")
    parser.add_argument("-i","--iso", default=200, type=int, metavar="")
    parser.add_argument("-f","--framerate", default=20, type=int, metavar="")
    parser.add_argument("-r","--res", default=(1640,1232), metavar="")
    parser.add_argument("-v","--vidsize", default=0.4, type=float, metavar="")

    args = parser.parse_args()
    res = tuple(args.res)
    Camconfig(auto = args.auto, iso = args.iso, framerate = args.framerate,
              res = res, vidsize = args.vidsize)
